/* libRNGnet template classes library, Version 1.0.
 * 
 * Copyright Vladimir Shakhov 2011
 * Distributed under the rules of SB RAS Algorithms&Programmes Fund, 
 * for details see http://fap.sbras.ru
 *
 * The library of pseudorandom number generators for networks fault tolerance analysis.  
 * The library is developed within a project supported by the Russian Foundation for Basic Research (grant 11-07-00183)
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT 
 * SHALL THE AUTHOR  OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE FOR ANY DAMAGES, CLAIMS 
 * OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 * 
 * 
 * Revision history
 * 2011-11-11 a set of modified exponential distributions are added
 * Contact persons: shakhov@rav.sscc.ru
 */

#ifndef __LIBRNGNET_H__
#define __LIBRNGNET_H__

#include <random> 


/** 
***  generator for Weibull distribution 
**/
template<typename RealType = double, typename IntType = int>
    class weibull_distribution {
private:
      RealType lambda1;
	  IntType k1;

public:
    typedef RealType input_type;
	typedef IntType shape_type;
    typedef RealType result_type;
    
	explicit weibull_distribution(const result_type& lambda0 = result_type(1), const shape_type& k0 = shape_type(1) )
	{
		lambda1 = lambda0;
		k1 = k0;
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
	result_type shape() const
	{
		return k1;
	}

	void reset(){}

	template<typename BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
		return ( pow( -log(_NRAND(bG_, RealType)), result_type(1)/result_type(k1) ) * lambda1);        
		}
 };

/** 
***  generator for Pareto distribution 
**/
template<typename RealType = double>
    class pareto_distribution {
private:
      RealType lambda1;
	  RealType k1;

public:
    typedef RealType input_type;
	typedef RealType result_type;
    
	explicit pareto_distribution(const result_type& lambda0 = result_type(1), const result_type& k0 = result_type(1) )
	{
		lambda1 = lambda0;
		k1 = k0;
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
	result_type shape() const
	{
		return k1;
	}

	void reset(){}

	template<typename BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
		return ( lambda1 / pow( _NRAND(bG_, RealType), result_type(1)/result_type(k1)) );        
		}
 };


/** 
***  generator for shifted exponential distribution 
**/
template<typename RealType = double>
    class shifted_exponential
{
	private:
		RealType lambda1;
	    RealType shift1;

	public:
		typedef RealType input_type;
		typedef RealType result_type;
    
	explicit shifted_exponential(const input_type& lambda0 = result_type(1), const input_type& shift0 = result_type(0))
	{
		lambda1 = lambda0;
		shift1 = shift0;
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
    result_type shift() const
	{
		return shift1;
	}

	result_type mean() const
	{
		return result_type(1)/lambda1 + shift1;
	}

	void reset(){}

	template<typename BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
		return (shift1 - log(_NRAND(bG_, RealType)) / lambda1);        
		}
 }; // template shifted_exponential


/** 
***  generator for shifted exponential distribution 
**/
template<typename RealType = double>
    class truncated_exponential 
{
	private:
		RealType lambda1;
		RealType left_lim1;
		RealType twait;

	public:
		typedef RealType input_type;
		typedef RealType result_type;
    
	explicit truncated_exponential(const result_type& lambda0 = result_type(1), const result_type& left_lim0 = result_type(1))
	{
		lambda1 = lambda0;
		left_lim1 = left_lim0;		
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
    result_type left_lim() const
	{
		return left_lim11;
	}

	result_type mean() const
	{
		return result_type(1)/lambda1 - left_lim1/(exp(left_lim1*lambda1)-1);
	}

	void reset(){}

	template<class BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
			twait = left_lim1 + 1.0;
			while (twait > left_lim1) {twait = (- log(_NRAND(bG_, RealType)) / lambda1);}
		return twait;        
		} //operator()
 }; // SYN_exponential


/** 
***  Generator for  waiting times of SYN packets in backlog queue. 
***  It is a composite distribution.
**/

template<typename RealType = double>
    class SYN_exponential 
{
	private:
		RealType lambda1;
		RealType timeout1;
		RealType twait;

	public:
		typedef RealType input_type;
		typedef RealType result_type;
    
	explicit SYN_exponential(const result_type& lambda0 = result_type(1), const result_type& timeout0 = result_type(1))
	{
		lambda1 = lambda0;
		timeout1 = timeout0;		
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
    result_type timeout() const
	{
		return timeout1;
	}

	result_type mean() const
	{
		return result_type(1)/lambda1 * (result_type(1) - exp(-timeout1*lambda1));
	}

	void reset(){}

	template<class BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
			twait = (- log(_NRAND(bG_, RealType)) / lambda1);
			if (twait > timeout1) twait = timeout1;
			return twait;        
		} //operator()
 }; // SYN_exponential


template<typename RealType = double>
    class pfSYN_exponential 
{
	private:
		RealType lambda1;
		RealType timeout1;
		RealType pf1;
		RealType twait;

	public:
		typedef RealType input_type;
		typedef RealType result_type;
    
	explicit pfSYN_exponential(const result_type& lambda0 = result_type(1), const result_type& timeout0 = result_type(1), const result_type& pf0 = result_type(0))
	{
		lambda1 = lambda0;
		timeout1 = timeout0;
		pf1 = pf0;
	}

	result_type lambda() const
	{
		return lambda1;
	}
    
    result_type timeout() const
	{
		return timeout1;
	}
    
    result_type prob_filter() const
	{
		return pf1;
	}

	result_type mean() const
	{
		return (1-pf1)*(result_type(1)/lambda1 * (result_type(1) - exp(-timeout1*lambda1)));
	}

	void reset(){}

	template<class BaseGenerator>
		result_type operator()(BaseGenerator& bG_)
		{	
			if (_NRAND(bG_, RealType) < pf1) twait = result_type(0);
			else
			{ 
				twait = (- log(_NRAND(bG_, RealType)) / lambda1);
				if (twait > timeout1) twait = timeout1;
			}
			return twait;        
		} //operator()
 }; // SYN_exponential


#endif // __LIBRNGNET_H__